"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const tsutils_1 = require("tsutils");
const typescript_1 = __importDefault(require("typescript"));
/**
 * An ESLint rule to ban unsafe usage of the array/object index operator, which is not well-typed in TypeScript.
 * @see https://github.com/Microsoft/TypeScript/issues/13778
 */
const noArraySubscript = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Array and object subscript access are not type-safe in TypeScript.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringGeneric: "This subscript access is not type-safe in TypeScript.",
        },
        schema: [],
    },
    create: (context) => {
        // eslint-disable-next-line total-functions/no-unsafe-assignment
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const reportUnsafeSubscriptAccess = (node
        // eslint-disable-next-line functional/no-return-void, sonarjs/cognitive-complexity
        ) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (!node.computed) {
                // Allow non-computed (regular) property access.
                return;
            }
            const contextualType = checker.getContextualType(parserServices.esTreeNodeToTSNodeMap.get(node));
            // eslint-disable-next-line functional/no-conditional-statement
            if (contextualType !== undefined &&
                tsutils_1.unionTypeParts(contextualType).some((t) => t.flags & typescript_1.default.TypeFlags.Undefined)) {
                // Contextual type includes (or is) undefined.
                // See https://www.typescriptlang.org/docs/handbook/type-inference.html#contextual-typing
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statement
            if (node.parent !== undefined &&
                node.parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression &&
                node.parent.left === node) {
                // This is the left hand side of an array assignment, so there's no partiality issue here.
                // eslint-plugin-functional will catch the mutation.
                return;
            }
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node.object);
            const type = checker.getTypeAtLocation(tsNode);
            const numberIndexType = type.getNumberIndexType();
            // eslint-disable-next-line functional/no-conditional-statement
            if (numberIndexType !== undefined &&
                tsutils_1.unionTypeParts(numberIndexType).some((t) => t.flags & typescript_1.default.TypeFlags.Undefined)) {
                // Allow subscript access if undefined is already in the array type.
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statement
            if (tsutils_1.isTupleTypeReference(type) &&
                node.property.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                typeof node.property.value === "number" &&
                node.property.value < type.target.minLength) {
                // Allow tuples (with or without a ...rest element) if
                // the index we're accessing is known to be safe at compile time.
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statement
            if (tsutils_1.isTupleTypeReference(type) && type.target.hasRestElement === false) {
                // Always allow tuples without a ...rest element.
                // If this access is invalid TypeScript itself will catch it.
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statement
            if (type.getStringIndexType() === undefined &&
                type.getNumberIndexType() === undefined) {
                // Allow object subscript access when there is no index signature in the object (i.e. when this object is not a Record<string, B> or a Record<number, B>).
                // If this access is invalid TypeScript itself will catch it.
                return;
            }
            // eslint-disable-next-line functional/no-expression-statement
            context.report({
                node: node,
                messageId: "errorStringGeneric",
            });
        };
        return {
            // eslint-disable-next-line functional/no-return-void
            OptionalMemberExpression: (node) => {
                return reportUnsafeSubscriptAccess(node);
            },
            // eslint-disable-next-line functional/no-return-void
            MemberExpression: (node) => {
                return reportUnsafeSubscriptAccess(node);
            },
        };
    },
};
exports.default = noArraySubscript;
